// Sugarscape in Swarm. Copyright  1997 Nelson Minar
// Class Predator based on SugarAgent Copyright @ 2004 Paul Johnson
// This program is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

#import "Predator.h"
#import <simtools.h>
#import "ModelSwarm.h"
#import <gui.h>

@implementation Predator


- createEnd
{
  numKilled = 0;
  return [super createEnd];
}


// One "step" for an agent. Depending on the rules in effect, this step
//   could have a lot of different meanings.
- step
{

  age++;

  SugarAgent * targetAgent = [self move];

  currentSugar += [targetAgent getCurrentSugar];

  currentSugar -= metabolism;

  if (targetAgent)
    {
#ifdef PREDEBUG
      fprintf(stderr,"I'm at %d,%d, Killing one at %d %d sugar= %d\n",x,y,targetAgent->x,targetAgent->y,[targetAgent getCurrentSugar]);
#endif
      [sugarSpace removeAgent: targetAgent];
      [modelSwarm agentDeath: targetAgent];
      numKilled++;
    }
  // Check if I'm dying
  // if (currentSugar <= 0 || age >= deathAge)
  //  {
  //    [sugarSpace removeAgent: self];
  //    [modelSwarm agentDeath: self];
  //  }
  
  return self;
}


// A dumb macro to get the absolute value of an integer
#define intabs(a) ((a) < 0 ? -(a) : (a))

- move
{
  int xLook, yLook;
  SugarValue bestSugar;
  int bestDistance;
  int goodSpots;
  int goodX[16], goodY[16];			  // 4 should be adequate
  int chosenSpot, newX, newY;
  SugarAgent * targetAgent = nil;
  int targetSugar;
  
  // prime the algorithm with out of range values
  bestSugar = -1;
  goodSpots = 0;
  bestDistance = 999999;			  // big number

  // First, look in the X direction for good spots.
  yLook = y;
  for (xLook = x - vision; xLook <= x + vision; xLook++)
    {
      // Is the spot currently empty?
      if ((targetAgent = [sugarSpace getAgentAtX: xLook Y: yLook]) != nil)
        {
          // is the spot we're looking at the best we've ever seen?
          if ((targetSugar = [targetAgent getCurrentSugar]) > bestSugar)
            {
              // yes, best spot ever. Forget everything else, record this
              // as the only good spot
              bestSugar = targetSugar;
              bestDistance = intabs(x - xLook);
              goodSpots = 0;
              goodX[0] = xLook;
              goodY[0] = yLook;
              goodSpots++;
            }
          else if (targetSugar == bestSugar)
            {
              // No, it's only as good as anything else we've seen. Is it closer
              // than any other spot we've seen with this sugar?
              if (intabs(x - xLook) < bestDistance) {
                // Yes, forget all the rest - this is the only good spot
                bestDistance = intabs(x - xLook);
                goodSpots = 0;
                goodX[0] = xLook;
                goodY[0] = yLook;
                goodSpots++;
              }
              else if (intabs (x - xLook) == bestDistance)
                {
                  // No, this spot is as good as some other one. Just add this
                  // one on as a good spot.
                  goodX[goodSpots] = xLook;
                  goodY[goodSpots] = yLook;
                  goodSpots++;
                }
            } 
        }
    }
  
  // Now repeat the same choice in the Y axis, ignoring current y position.
  xLook = x;
  for (yLook = y - vision; yLook <= y + vision; yLook++)
    {
      if ((yLook != y) && ((targetAgent = [sugarSpace getAgentAtX: xLook Y: yLook]) != nil))
        {
          if ((targetSugar = [targetAgent getCurrentSugar]) > bestSugar)
            {
              bestSugar = targetSugar;
              bestDistance = intabs(y - yLook);
              goodSpots = 0;
              goodX[0] = xLook;
              goodY[0] = yLook;
              goodSpots++;
            }
          else if (targetSugar == bestSugar)
            {
              if (intabs(y - yLook) < bestDistance)
                {
                  bestDistance = intabs(y - yLook);
                  goodSpots = 0;
                  goodX[0] = xLook;
                  goodY[0] = yLook;
                  goodSpots++;
                }
              else if (intabs(y - yLook) == bestDistance)
                {
                  goodX[goodSpots] = xLook;
                  goodY[goodSpots] = yLook;
                  goodSpots++;
                }
            } 
        }
    }
  
  // A bit of debug printing to make sure the agents are behaving sensibly.
  // (turned off normally)
#ifdef PREDEBUG
  {
    int i;
    fprintf(stderr,"Predator at %d,%d Found %d good spots:\n", x,y, goodSpots);
    for (i = 0; i < goodSpots; i++)
      fprintf(stderr,"  (%d,%d) = %d\n", goodX[i], goodY[i],
	    [ [sugarSpace getAgentAtX: goodX[i] Y: goodY[i]] getCurrentSugar] );
  }
#endif
  
  // Alright, goodX[] and goodY[] record the best spots we've found.
  // Let's figure out where to move.

  if (goodSpots == 0)				  // No spots are good
    ;						  // don't even move
  else
    {
      if (goodSpots == 1)				  // only one good spot
        chosenSpot = 0;
      else					  // pick a random spot
        chosenSpot = [uniformIntRand getIntegerWithMin: 0 withMax: goodSpots-1];
      newX = goodX[chosenSpot];			  // get the coordinate
      newY = goodY[chosenSpot];			  // and move there!
      

      [sugarSpace movePredator: self toX: newX Y: newY];
#ifdef PREDEBUG
             
      fprintf(stderr,"Predator moved to %d,%d:\n", x,y);
#endif
    }
  return targetAgent = [sugarSpace getAgentAtX: x Y: y];
}

//// The code below here is not very interesting - it has to do with
//// the technical details of managing the data in the model, not the
//// modelling itself. You can safely ignore it until you want to know
//// the details of how the program works.

// set methods for various values
- setModelSwarm: s
{
  modelSwarm = s;
  sugarSpace = [s getSugarSpace];
  return self;
}

- (SugarValue)getCurrentSugar
{
  return currentSugar;
}

- setCurrentSugar: (SugarValue)cs
{
  currentSugar = cs;
  return self;
}

- (int)getMetabolism
{
  return metabolism;
}

- setMetabolism: (int)m
{
  metabolism = m;
  return self;
}

- (int)getVision
{
  return vision;
}

- setVision: (int)v
{
  vision = v;
  return self;
}

- (int)getAge
{
  return age;
}

- (int)getNumKilled
{
  return numKilled;
}

- setDeathAge: (int)s
{
  deathAge = s;
  return self;
}

// Graphics code - how to draw oneself (hardcoded colour value here. If
// agents have different properties, these colour should be different.)
- drawSelfOn: (id <Raster>)r
{
  // [r drawPointX: x Y: y Color: 100];
  if (myPixmap == nil)
    {
      myPixmap = [Pixmap createBegin: [self getZone]];
      [myPixmap setFile: "pacman.png"];
      myPixmap = [myPixmap createEnd];
    }

  [myPixmap setRaster: r];

  [r draw:  myPixmap  X: (int) x  Y: y];

  return self;
}

@end
